"""
DIALS Regression Data Manager
https://github.com/dials/data
"""

from __future__ import annotations

__all__: list = []
__author__ = """Markus Gerstel"""
__email__ = "dials-support@lists.sourceforge.net"
__version__ = "2.4.34"
__commit__ = "31eed6b0d27357aa201069108f6a272e7841f576"
__version_tuple__ = tuple(int(x) for x in __version__.split("."))
